# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
SQLAlchemy ORM model for the analysis run storage database.
"""
from datetime import datetime, timedelta
from math import ceil
import os
from typing import Optional

from sqlalchemy import Boolean, Column, DateTime, Enum, ForeignKey, Integer, \
    LargeBinary, MetaData, String, UniqueConstraint, Table, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql.expression import true, false

CC_META = MetaData(naming_convention={
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
})

# Create base class for ORM classes.
Base = declarative_base(metadata=CC_META)


class Checker(Base):
    """
    Records of a look-up table that associates a product-global ID for each
    analyzer name and checker name encountered.
    """
    __tablename__ = "checkers"

    id = Column(Integer, autoincrement=True, primary_key=True)
    analyzer_name = Column(String)
    checker_name = Column(String)
    severity = Column(Integer, index=True)

    __table_args__ = (
        UniqueConstraint("analyzer_name", "checker_name"),
    )

    def __init__(self, analyzer_name: str, checker_name: str, severity: int):
        self.analyzer_name = analyzer_name
        self.checker_name = checker_name
        self.severity = severity


class AnalysisInfoChecker(Base):
    __tablename__ = "analysis_info_checkers"

    analysis_info_id = Column(Integer,
                              ForeignKey("analysis_info.id",
                                         deferrable=True,
                                         initially="DEFERRED",
                                         ondelete="CASCADE"),
                              primary_key=True)
    checker_id = Column(Integer,
                        ForeignKey("checkers.id",
                                   deferrable=True,
                                   initially="DEFERRED",
                                   ondelete="RESTRICT"),
                        primary_key=True)
    enabled = Column(Boolean)

    def __init__(self,
                 analysis_info: "AnalysisInfo",
                 checker: Checker,
                 is_enabled: bool):
        self.analysis_info_id = analysis_info.id
        self.checker_id = checker.id
        self.enabled = is_enabled


class AnalysisInfo(Base):
    __tablename__ = "analysis_info"

    id = Column(Integer, autoincrement=True, primary_key=True)
    analyzer_command = Column(LargeBinary)
    available_checkers = relationship(AnalysisInfoChecker, uselist=True)

    def __init__(self, analyzer_command: bytes):
        self.analyzer_command = analyzer_command


class Run(Base):
    __tablename__ = 'runs'

    __table_args__ = (
        UniqueConstraint('name'),
    )

    id = Column(Integer, autoincrement=True, primary_key=True)
    date = Column(DateTime)
    duration = Column(Integer)  # Seconds, -1 if unfinished.
    name = Column(String)
    version = Column(String)
    can_delete = Column(Boolean, nullable=False, server_default=true(),
                        default=True)

    def __init__(self, name, version):
        self.date, self.name, self.version = datetime.now(), name, version
        self.duration = -1

    def mark_finished(self):
        if self.duration == -1:
            self.duration = ceil((datetime.now() - self.date).total_seconds())


class RunLock(Base):
    """
    Represents a lock record for a particular run name, constituting that the
    run identified by said name should NOT be stored into, as it is undergoing
    a write operation.
    """

    __tablename__ = 'run_locks'

    name = Column(String, nullable=False, primary_key=True)
    locked_at = Column(DateTime, nullable=False)
    username = Column(String, nullable=True)

    def __init__(self, run_name, username=None):
        """Create a new lock for the given run name."""
        self.name = run_name
        self.locked_at = datetime.now()
        self.username = username

    def touch(self):
        """Update the lock's timestamp to be the current one."""
        self.locked_at = datetime.now()

    def when_expires(self, grace_seconds):
        """Calculates when the current lock will expire assuming the
        expiration time is grace_seconds, and the lock will never be touched
        until this moment."""
        return self.locked_at + timedelta(seconds=grace_seconds)

    def has_expired(self, grace_seconds):
        """Returns if the lock has expired, i.e. since the last touch()
        or creation, grace_seconds number of seconds has passed."""
        return datetime.now() > self.when_expires(grace_seconds)


class AnalyzerStatistic(Base):
    __tablename__ = 'analyzer_statistics'

    id = Column(Integer, autoincrement=True, primary_key=True)
    run_history_id = Column(Integer,
                            ForeignKey('run_histories.id',
                                       deferrable=True,
                                       initially="DEFERRED",
                                       ondelete='CASCADE'),
                            index=True)
    analyzer_type = Column(String)
    version = Column(LargeBinary)
    successful = Column(Integer)
    failed = Column(Integer)
    failed_files = Column(LargeBinary, nullable=True)

    def __init__(self, run_history_id, analyzer_type, version, successful,
                 failed, failed_files):
        self.run_history_id = run_history_id
        self.analyzer_type = analyzer_type
        self.version = version
        self.successful = successful
        self.failed = failed
        self.failed_files = failed_files


RunHistoryAnalysisInfo = Table(
    'run_history_analysis_info',
    Base.metadata,
    Column(
        'run_history_id',
        Integer,
        ForeignKey('run_histories.id',
                   deferrable=True,
                   initially="DEFERRED",
                   ondelete="CASCADE"),
        index=True),
    Column('analysis_info_id', Integer, ForeignKey('analysis_info.id'))
)


class RunHistory(Base):
    __tablename__ = 'run_histories'

    id = Column(Integer, autoincrement=True, primary_key=True)
    run_id = Column(Integer,
                    ForeignKey('runs.id', deferrable=True,
                               initially="DEFERRED", ondelete='CASCADE'),
                    index=True)
    version_tag = Column(String)
    user = Column(String, nullable=False)
    time = Column(DateTime, nullable=False)
    cc_version = Column(String, nullable=True)
    description = Column(String, nullable=True)

    run = relationship(Run, uselist=False)
    analyzer_statistics = relationship(AnalyzerStatistic,
                                       lazy="joined")

    analysis_info = relationship(
        "AnalysisInfo",
        secondary=RunHistoryAnalysisInfo)

    __table_args__ = (UniqueConstraint('run_id', 'version_tag'),)

    def __init__(self, run_id, version_tag, user, time, cc_version,
                 description):
        self.run_id = run_id
        self.version_tag = version_tag
        self.user = user
        self.time = time
        self.cc_version = cc_version
        self.description = description


class FileContent(Base):
    __tablename__ = 'file_contents'

    content_hash = Column(String, primary_key=True)
    content = Column(LargeBinary)

    # Note: two different authors can commit the same file content to
    # different paths in which case the blame info will be the same.
    blame_info = Column(LargeBinary, nullable=True)

    def __init__(self, content_hash, content, blame_info):
        self.content_hash, self.content, self.blame_info = \
            content_hash, content, blame_info


class File(Base):
    __tablename__ = 'files'

    id = Column(Integer, autoincrement=True, primary_key=True)
    filepath = Column(String, nullable=False)
    filename = Column(String, nullable=False)
    content_hash = Column(String,
                          ForeignKey('file_contents.content_hash',
                                     deferrable=True,
                                     initially="DEFERRED", ondelete='CASCADE'),
                          index=True)
    remote_url = Column(String, nullable=True)
    tracking_branch = Column(String, nullable=True)

    __table_args__ = (UniqueConstraint('filepath', 'content_hash'),)

    def __init__(self, filepath, content_hash, remote_url, tracking_branch):
        self.filepath = filepath
        self.filename = os.path.basename(filepath)
        self.content_hash = content_hash
        self.remote_url = remote_url
        self.tracking_branch = tracking_branch


class BugPathEvent(Base):
    __tablename__ = 'bug_path_events'

    line_begin = Column(Integer)
    col_begin = Column(Integer)
    line_end = Column(Integer)
    col_end = Column(Integer)

    order = Column(Integer, primary_key=True)

    msg = Column(String)
    file_id = Column(Integer, ForeignKey('files.id', deferrable=True,
                                         initially="DEFERRED",
                                         ondelete='CASCADE'), index=True)
    report_id = Column(Integer, ForeignKey('reports.id', deferrable=True,
                                           initially="DEFERRED",
                                           ondelete='CASCADE'),
                       index=True,
                       primary_key=True)

    def __init__(self, line_begin, col_begin, line_end, col_end,
                 order, msg, file_id, report_id):
        self.line_begin, self.col_begin, self.line_end, self.col_end = \
            line_begin, col_begin, line_end, col_end

        self.order = order
        self.msg = msg
        self.file_id = file_id
        self.report_id = report_id


class BugReportPoint(Base):
    __tablename__ = 'bug_report_points'

    line_begin = Column(Integer)
    col_begin = Column(Integer)
    line_end = Column(Integer)
    col_end = Column(Integer)

    order = Column(Integer, primary_key=True)

    file_id = Column(Integer, ForeignKey('files.id', deferrable=True,
                                         initially="DEFERRED",
                                         ondelete='CASCADE'), index=True)
    report_id = Column(Integer, ForeignKey('reports.id', deferrable=True,
                                           initially="DEFERRED",
                                           ondelete='CASCADE'),
                       index=True,
                       primary_key=True)

    def __init__(self, line_begin, col_begin, line_end, col_end,
                 order, file_id, report_id):
        self.line_begin, self.col_begin, self.line_end, self.col_end = \
            line_begin, col_begin, line_end, col_end

        self.order = order
        self.file_id = file_id
        self.report_id = report_id


class ExtendedReportData(Base):
    """
    Store extra information which can help to understand or fix a report.
    """
    __tablename__ = 'extended_report_data'

    id = Column(Integer, autoincrement=True, primary_key=True)

    report_id = Column(Integer, ForeignKey('reports.id', deferrable=True,
                                           initially="DEFERRED",
                                           ondelete='CASCADE'),
                       index=True)

    file_id = Column(Integer, ForeignKey('files.id', deferrable=True,
                                         initially="DEFERRED",
                                         ondelete='CASCADE'), index=True)

    type = Column(Enum('note',
                       'macro',
                       'fixit',
                       name='extended_data_type'))

    line_begin = Column(Integer)
    col_begin = Column(Integer)
    line_end = Column(Integer)
    col_end = Column(Integer)

    message = Column(String)

    def __init__(self, line_begin, col_begin, line_end, col_end,
                 message, file_id, report_id, data_type):

        self.line_begin = line_begin
        self.col_begin = col_begin
        self.line_end = line_end
        self.col_end = col_end
        self.message = message
        self.file_id = file_id
        self.report_id = report_id
        self.type = data_type


ReportAnalysisInfo = Table(
    'report_analysis_info',
    Base.metadata,
    Column(
        'report_id',
        Integer,
        ForeignKey('reports.id',
                   deferrable=True,
                   initially="DEFERRED",
                   ondelete="CASCADE"),
        index=True),
    Column('analysis_info_id', Integer, ForeignKey('analysis_info.id'))
)


ReviewStatusType = Enum(
    'unreviewed',
    'confirmed',
    'false_positive',
    'intentional',
    name='review_status')


class Report(Base):
    __tablename__ = 'reports'

    id = Column(Integer, autoincrement=True, primary_key=True)
    file_id = Column(Integer, ForeignKey('files.id', deferrable=True,
                                         initially="DEFERRED",
                                         ondelete='CASCADE'),
                     index=True)
    run_id = Column(Integer,
                    ForeignKey('runs.id', deferrable=True,
                               initially="DEFERRED",
                               ondelete='CASCADE'),
                    index=True)
    bug_id = Column(String, index=True)
    checker_id = Column(Integer, ForeignKey("checkers.id",
                                            deferrable=False,
                                            ondelete="RESTRICT"),
                        nullable=False,
                        index=True)
    checker = relationship(Checker, innerjoin=True, lazy="joined",
                           foreign_keys=[checker_id])

    line = Column(Integer)
    column = Column(Integer)
    path_length = Column(Integer)

    # TODO: multiple messages to multiple source locations?
    checker_message = Column(String)
    detection_status = Column(Enum('new',
                                   'unresolved',
                                   'resolved',
                                   'reopened',
                                   'off',
                                   'unavailable',
                                   name='detection_status'))
    review_status = Column(ReviewStatusType,
                           nullable=False,
                           server_default='unreviewed')
    review_status_author = Column(String)
    review_status_message = Column(LargeBinary)
    review_status_date = Column(DateTime, nullable=True)
    # We'd like to indicate whether a suppression comes from a source code
    # comment or set via the GUI. Former ones must not change when set from
    # GUI.
    review_status_is_in_source = Column(
        Boolean, nullable=False, server_default=false())

    detected_at = Column(DateTime, nullable=False)

    # A report is considered as "fixed" when it is not found in the project
    # anymore either based on its detection status or its review status is set
    # to false positive or intentional.
    fixed_at = Column(DateTime)

    analysis_info = relationship(
        "AnalysisInfo",
        secondary=ReportAnalysisInfo)

    # Cascade delete might remove rows, SQLAlchemy warns about this.
    # To remove warnings about already deleted items set this to False.
    __mapper_args__ = {
        'confirm_deleted_rows': False
    }

    annotations = relationship("ReportAnnotations")

    def __init__(self,
                 file_id: int,
                 run_id: int,
                 bug_id: Optional[str],
                 checker: Checker,
                 line: int,
                 column: int,
                 path_length: int,
                 checker_message: str,
                 detection_status,
                 review_status,
                 review_status_author: Optional[str],
                 review_status_message: Optional[bytes],
                 review_status_date: Optional[datetime],
                 review_status_is_in_source: bool, detection_date: datetime,
                 fixed_date: Optional[datetime]):
        self.file_id = file_id
        self.run_id = run_id
        self.bug_id = bug_id
        self.checker = checker
        self.line = line
        self.column = column
        self.path_length = path_length
        self.checker_message = checker_message
        self.detection_status = detection_status
        self.review_status = review_status
        self.review_status_author = review_status_author
        self.review_status_message = review_status_message
        self.review_status_date = review_status_date
        self.review_status_is_in_source = review_status_is_in_source
        self.detected_at = detection_date
        self.fixed_at = fixed_date


class ReportAnnotations(Base):
    __tablename__ = "report_annotations"

    def __init__(self, report_id: int, key: str, value: str):
        self.report_id = report_id
        self.key = key
        self.value = value

    report_id = Column(
        Integer,
        ForeignKey("reports.id", ondelete="CASCADE"),
        primary_key=True,
        index=True)
    key = Column(String, primary_key=True, nullable=False)
    value = Column(String, nullable=False)


class Comment(Base):
    __tablename__ = 'comments'

    id = Column(Integer, autoincrement=True, primary_key=True)
    bug_hash = Column(String, nullable=False, index=True)
    author = Column(String, nullable=False)
    message = Column(LargeBinary, nullable=False)

    # Default value is 0 which means a user given comment.
    kind = Column(Integer,
                  nullable=False,
                  server_default="0")

    created_at = Column(DateTime, nullable=False)

    def __init__(self, bug_hash, author, message, kind, created_at):
        self.bug_hash = bug_hash
        self.author = author
        self.message = message
        self.kind = kind
        self.created_at = created_at


class ReviewStatus(Base):
    """
    This table contains a mapping from bug hashes to review statuses. This
    mapping determines the default review status of a newly added report which
    doesn't have a review status source code comment. In case of no source code
    comment and no entry in this table the default review status is
    "unreviewed".
    """
    __tablename__ = 'review_statuses'

    bug_hash = Column(String, primary_key=True)
    status = Column(ReviewStatusType, nullable=False)
    author = Column(String, nullable=False)
    message = Column(LargeBinary, nullable=False)
    date = Column(DateTime, nullable=False)


class SourceComponent(Base):
    __tablename__ = 'source_components'

    name = Column(String, nullable=False, primary_key=True)

    # Contains multiple file paths separated by new line characters. Each file
    # path start with a '+' (path should be filtered) or '-' (path should not
    # be filtered) sign. E.g.: "+/a/b/x.cpp\n-/a/b/"
    value = Column(LargeBinary, nullable=False)

    description = Column(Text, nullable=True)
    username = Column(String, nullable=True)

    def __init__(self, name, value, description=None, user_name=None):
        self.name = name
        self.value = value
        self.description = description
        self.username = user_name


class CleanupPlan(Base):
    __tablename__ = 'cleanup_plans'

    __table_args__ = (
        UniqueConstraint('name'),
    )

    id = Column(Integer, autoincrement=True, primary_key=True)
    name = Column(String, nullable=False)
    due_date = Column(DateTime, nullable=True)
    description = Column(String, nullable=True)
    closed_at = Column(DateTime, nullable=True)

    def __init__(self, name, due_date=None, description=None, closed_at=None):
        self.name = name
        self.due_date = due_date
        self.description = description
        self.closed_at = closed_at


class CleanupPlanReportHash(Base):
    __tablename__ = 'cleanup_plan_report_hashes'

    cleanup_plan_id = Column(
        Integer,
        ForeignKey('cleanup_plans.id',
                   deferrable=True,
                   initially="DEFERRED",
                   ondelete="CASCADE"),
        index=True)

    bug_hash = Column(String, primary_key=True)


IDENTIFIER = {
    'identifier': "RunDatabase",
    'orm_meta': CC_META
}



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Handler for the subcommand that is used to start and manage CodeChecker
servers, which are used to query analysis report information.
"""


import argparse
import errno
from functools import partial
import os
import signal
import socket
import sys
import time
from typing import List, Optional, Tuple, cast

from alembic import config
from alembic import script
from alembic.util import CommandError
import psutil
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import sessionmaker

from codechecker_api_shared.ttypes import DBStatus

from codechecker_report_converter import twodim

from codechecker_common import arg, cmd_config, logger, util
from codechecker_common.compatibility.multiprocessing import Pool, cpu_count

from codechecker_server import instance_manager, server
from codechecker_server.database import database
from codechecker_server.database.config_db_model \
    import IDENTIFIER as CONFIG_META
from codechecker_server.database.config_db_model \
    import Product as ORMProduct
from codechecker_server.database.run_db_model \
    import IDENTIFIER as RUN_META

from codechecker_web.shared import webserver_context, database_status, \
    host_check, env

LOG = logger.get_logger('server')


def get_argparser_ctor_args():
    """
    This method returns a dict containing the kwargs for constructing an
    argparse.ArgumentParser (either directly or as a subparser).
    """

    return {
        'prog': 'CodeChecker server',
        'formatter_class': arg.RawDescriptionDefaultHelpFormatter,

        # Description is shown when the command's help is queried directly
        'description': """
The CodeChecker Web server is used to handle the storage and navigation of
analysis results. A started server can be connected to via a Web browser, or
by using the 'CodeChecker cmd' command-line client.""",

        # Help is shown when the "parent" CodeChecker command lists the
        # individual subcommands.
        'help': "Start and manage the CodeChecker Web server."
    }


def add_arguments_to_parser(parser):
    """
    Add the subcommand's arguments to the given argparse.ArgumentParser.
    """

    default_workspace = env.get_default_workspace()

    # TODO: --workspace is an outdated concept in 'store'. Later on,
    # it shall be deprecated, as changes to db_handler commence.
    parser.add_argument('-w', '--workspace',
                        type=str,
                        dest="workspace",
                        default=default_workspace,
                        required=False,
                        help="Directory where CodeChecker can store analysis "
                             "result related data, such as the database. "
                             "(Cannot be specified at the same time with "
                             "'--sqlite' or '--config-directory'.)")

    parser.add_argument('-f', '--config-directory',
                        type=str,
                        dest="config_directory",
                        default=default_workspace,
                        required=False,
                        help="Directory where CodeChecker server should read "
                             "server-specific configuration (such as "
                             "authentication settings, TLS certificate"
                             " (cert.pem) and key (key.pem)) from.")

    parser.add_argument('--host',
                        type=str,
                        dest="listen_address",
                        default="localhost",
                        required=False,
                        help="The IP address or hostname of the server on "
                             "which it should listen for connections. "
                             "For IPv6 listening, specify an IPv6 address, "
                             "such as \"::1\".")

    # TODO: -v/--view-port is too verbose. The server's -p/--port is used
    # symmetrically in 'CodeChecker cmd' anyways.
    parser.add_argument('-v', '--view-port',  # TODO: <- Deprecate and remove.
                        '-p', '--port',
                        type=int,
                        dest="view_port",
                        metavar='PORT',
                        default=8001,
                        required=False,
                        help="The port which will be used as listen port for "
                             "the server.")

    # TODO: This should be removed later on, in favour of --host.
    parser.add_argument('--not-host-only',
                        dest="not_host_only",
                        action="store_true",
                        required=False,
                        help="If specified, storing and viewing the results "
                             "will be possible not only by browsers and "
                             "clients running locally, but to everyone, who "
                             "can access the server over the Internet. "
                             "(Equivalent to specifying '--host \"::\"'.)")

    parser.add_argument('--skip-db-cleanup',
                        dest="skip_db_cleanup",
                        action='store_true',
                        default=False,
                        required=False,
                        help="Skip performing cleanup jobs on the database "
                             "like removing unused files.")

    cmd_config.add_option(parser)

    dbmodes = parser.add_argument_group("configuration database arguments")

    dbmodes = dbmodes.add_mutually_exclusive_group(required=False)

    dbmodes.add_argument('--sqlite',
                         type=str,
                         dest="sqlite",
                         metavar='SQLITE_FILE',
                         default=os.path.join(
                             '<CONFIG_DIRECTORY>',
                             "config.sqlite"),
                         required=False,
                         help="Path of the SQLite database file to use.")

    dbmodes.add_argument('--postgresql',
                         dest="postgresql",
                         action='store_true',
                         required=False,
                         default=argparse.SUPPRESS,
                         help="Specifies that a PostgreSQL database is to be "
                              "used instead of SQLite. See the \"PostgreSQL "
                              "arguments\" section on how to configure the "
                              "database connection.")

    pgsql = parser.add_argument_group("PostgreSQL arguments",
                                      "Values of these arguments are ignored, "
                                      "unless '--postgresql' is specified!")

    # TODO: --dbSOMETHING arguments are kept to not break interface from
    # old command. Database using commands such as "CodeChecker store" no
    # longer supports these --- it would be ideal to break and remove args
    # with this style and only keep --db-SOMETHING.
    pgsql.add_argument('--dbaddress', '--db-host',
                       type=str,
                       dest="dbaddress",
                       default="localhost",
                       required=False,
                       help="Database server address.")

    pgsql.add_argument('--dbport', '--db-port',
                       type=int,
                       dest="dbport",
                       default=5432,
                       required=False,
                       help="Database server port.")

    pgsql.add_argument('--dbusername', '--db-username',
                       type=str,
                       dest="dbusername",
                       default='codechecker',
                       required=False,
                       help="Username to use for connection.")

    pgsql.add_argument('--dbname', '--db-name',
                       type=str,
                       dest="dbname",
                       default="config",
                       required=False,
                       help="Name of the database to use.")

    root_account = parser.add_argument_group(
        "root account arguments",
        """
Servers automatically create a root user to access the server's configuration
via the clients. This user is created at first start and saved in the
CONFIG_DIRECTORY, and the credentials are printed to the server's standard
output. The plaintext credentials are NEVER accessible again.""")

    root_account.add_argument('--reset-root',
                              dest="reset_root",
                              action='store_true',
                              default=argparse.SUPPRESS,
                              required=False,
                              help="Force the server to recreate the master "
                                   "superuser (root) account name and "
                                   "password. The previous credentials will "
                                   "be invalidated, and the new ones will be "
                                   "printed to the standard output.")

    root_account.add_argument('--force-authentication',
                              dest="force_auth",
                              action='store_true',
                              default=argparse.SUPPRESS,
                              required=False,
                              help="Force the server to run in "
                                   "authentication requiring mode, despite "
                                   "the configuration value in "
                                   "'server_config.json'. This is needed "
                                   "if you need to edit the product "
                                   "configuration of a server that would not "
                                   "require authentication otherwise.")

    instance_mgmnt = parser.add_argument_group("running server management")

    instance_mgmnt = instance_mgmnt. \
        add_mutually_exclusive_group(required=False)

    instance_mgmnt.add_argument('-l', '--list',
                                dest="list",
                                action='store_true',
                                default=argparse.SUPPRESS,
                                required=False,
                                help="List the servers that has been started "
                                     "by you.")

    instance_mgmnt.add_argument('-r', '--reload',
                                dest="reload",
                                action='store_true',
                                default=argparse.SUPPRESS,
                                required=False,
                                help="Sends the CodeChecker server process a "
                                     "SIGHUP signal, causing it to reread "
                                     "it's configuration files.")

    # TODO: '-s' was removed from 'quickcheck', it shouldn't be here either?
    instance_mgmnt.add_argument('-s', '--stop',
                                dest="stop",
                                action='store_true',
                                default=argparse.SUPPRESS,
                                required=False,
                                help="Stops the server associated with "
                                     "the given view-port and workspace.")

    instance_mgmnt.add_argument('--stop-all',
                                dest="stop_all",
                                action='store_true',
                                default=argparse.SUPPRESS,
                                required=False,
                                help="Stops all of your running CodeChecker "
                                     "server instances.")

    database_mgmnt = parser.add_argument_group(
            "Database management arguments.",
            """
WARNING these commands needs to be called with the same workspace and
configuration arguments as the server so the configuration database will be
found which is required for the schema migration. Migration can be done
without a running server but pay attention to use the same arguments which
will be used to start the server.

NOTE:
Before migration it is advised to create a full a backup of the product
databases.
""")

    database_mgmnt = database_mgmnt. \
        add_mutually_exclusive_group(required=False)

    database_mgmnt.add_argument('--db-status',
                                type=str,
                                dest="status",
                                action='store',
                                default=argparse.SUPPRESS,
                                required=False,
                                help="Name of the product to get "
                                     "the database status for. "
                                     "Use 'all' to list the database "
                                     "statuses for all of the products.")

    database_mgmnt.add_argument('--db-upgrade-schema',
                                type=str,
                                dest='product_to_upgrade',
                                action='store',
                                default=argparse.SUPPRESS,
                                required=False,
                                help="Name of the product to upgrade to the "
                                     "latest database schema available in "
                                     "the package. Use 'all' to upgrade all "
                                     "of the products. "
                                     "NOTE: Before migration it is advised"
                                     " to create a full backup of "
                                     "the product databases.")

    database_mgmnt.add_argument('--db-force-upgrade',
                                dest='force_upgrade',
                                action='store_true',
                                default=argparse.SUPPRESS,
                                required=False,
                                help="Force the server to do database "
                                     "migration without user interaction. "
                                     "NOTE: Please use with caution and "
                                     "before automatic migration it is "
                                     "advised to create a full backup of the "
                                     "product databases.")

    logger.add_verbose_arguments(parser)

    def __handle(args):
        """Custom handler for 'server' so custom error messages can be
        printed without having to capture 'parser' in main."""

        def arg_match(options):
            return util.arg_match(options, sys.argv[1:])

        # See if there is a "PostgreSQL argument" specified in the invocation
        # without '--postgresql' being there. There is no way to distinguish
        # a default argument and a deliberately specified argument without
        # inspecting sys.argv.
        options = ['--dbaddress', '--dbport', '--dbusername', '--dbname',
                   '--db-host', '--db-port', '--db-username', '--db-name']
        psql_args_matching = arg_match(options)
        if any(psql_args_matching) and\
                'postgresql' not in args:
            first_matching_arg = next(iter([match for match
                                            in psql_args_matching]))
            parser.error("argument {0}: not allowed without "
                         "argument --postgresql".format(first_matching_arg))
            # parser.error() terminates with return code 2.

        # --not-host-only is a "shortcut", actually a to-be-deprecated
        # call which means '--host ""'.
        # TODO: Actually deprecate --not-host-only later on.
        options = ['--not-host-only', '--host']
        if set(arg_match(options)) == set(options):
            parser.error("argument --not-host-only: not allowed with "
                         "argument --host, as it is a shortcut to --host "
                         "\"::\"")
        else:
            # Apply the shortcut.
            if arg_match(['--not-host-only']):
                args.listen_address = "::"  # Listen on every interface.

            # --not-host-only is just a shortcut optstring, no actual use
            # is intended later on.
            delattr(args, 'not_host_only')

        # --workspace and --sqlite cannot be specified either, as
        # both point to a database location.
        options = ['--sqlite', '--workspace']
        options_short = ['--sqlite', '-w']
        if set(arg_match(options)) == set(options) or \
                set(arg_match(options_short)) == set(options_short):
            parser.error("argument --sqlite: not allowed with "
                         "argument --workspace")

        # --workspace and --config-directory also aren't allowed together now,
        # the latter one is expected to replace the earlier.
        options = ['--config-directory', '--workspace']
        options_short = ['--config-directory', '-w']
        if set(arg_match(options)) == set(options) or \
                set(arg_match(options_short)) == set(options_short):
            parser.error("argument --config-directory: not allowed with "
                         "argument --workspace")

        # If workspace is specified, sqlite is workspace/config.sqlite
        # and config_directory is the workspace directory.
        if arg_match(['--workspace', '-w']):
            args.config_directory = args.workspace
            args.sqlite = os.path.join(args.workspace,
                                       'config.sqlite')
            setattr(args, 'dbdatadir', os.path.join(args.workspace,
                                                    'pgsql_data'))

        # Workspace should not exist as a Namespace key.
        delattr(args, 'workspace')

        if '<CONFIG_DIRECTORY>' in args.sqlite:
            # Replace the placeholder variable with the actual value.
            args.sqlite = args.sqlite.replace('<CONFIG_DIRECTORY>',
                                              args.config_directory)

        # Convert relative sqlite file path to absolute.
        if 'sqlite' in args:
            args.sqlite = os.path.abspath(args.sqlite)

        if 'postgresql' not in args:
            # Later called database modules need the argument to be actually
            # present, even though the default is suppressed in the optstring.
            setattr(args, 'postgresql', False)

            # This is not needed by the database starter as we are
            # running SQLite.
            if 'dbdatadir' in args:
                delattr(args, 'dbdatadir')
        else:
            # If --postgresql is given, --sqlite is useless.
            delattr(args, 'sqlite')

        # Indicate in args that we are in instance manager mode.
        if "list" in args or "stop" in args or "stop_all" in args:
            setattr(args, "instance_manager", True)

        # If everything is fine, do call the handler for the subcommand.
        main(args)

    parser.set_defaults(
        func=__handle, func_process_config_file=cmd_config.process_config_file)


def print_prod_status(prod_status):
    """
    Print the database statuses for each of the products.
    """

    header = ['Product endpoint', 'Database status',
              'Database location',
              'Schema version in the database',
              'Schema version in the package']
    rows = []

    for k, v in prod_status.items():
        db_status, schema_ver, package_ver, db_location = v
        db_status_msg = database_status.db_status_msg.get(db_status)
        if schema_ver == package_ver:
            schema_ver += " (up to date)"
        rows.append([k, db_status_msg, db_location, str(schema_ver),
                     package_ver])

    prod_status = twodim.to_str('table',
                                header,
                                rows,
                                sort_by_column_number=0)
    LOG.info('Status of products:\n%s', prod_status)


def get_schema_version_from_package(migration_root):
    """
    Returns the latest schema version in the package.
    """

    cfg = config.Config()
    cfg.set_main_option("script_location", migration_root)
    pckg_schema_ver = script.ScriptDirectory.from_config(cfg)
    return pckg_schema_ver.get_current_head()


def check_product_db_status(cfg_sql_server, migration_root, environ):
    """
    Check the products for database statuses.

    :returns: dictionary of product endpoints with database statuses
    """
    engine = cfg_sql_server.create_engine()
    config_session = sessionmaker(bind=engine)
    sess = config_session()

    products: List[ORMProduct] = list()
    try:
        products = sess.query(ORMProduct) \
            .order_by(ORMProduct.endpoint.asc()) \
            .all()
    except Exception as ex:
        LOG.debug(ex)
        LOG.error("Failed to get product configurations from the database.")
        LOG.error("Please check your command arguments.")
        sys.exit(1)
    finally:
        # sys.exit raises SystemExit, which still performs finally clauses!
        sess.close()
        engine.dispose()

    package_schema = get_schema_version_from_package(migration_root)

    db_errors = [DBStatus.FAILED_TO_CONNECT,
                 DBStatus.MISSING,
                 DBStatus.SCHEMA_INIT_ERROR,
                 DBStatus.SCHEMA_MISSING]

    prod_status = {}
    for pd in products:
        db = database.SQLServer.from_connection_string(pd.connection,
                                                       pd.endpoint,
                                                       RUN_META,
                                                       migration_root,
                                                       interactive=False,
                                                       env=environ)
        db_location = db.get_db_location()

        try:
            status = db.connect()
            s_ver = db.get_schema_version()
            if s_ver in db_errors:
                s_ver = None
            prod_status[pd.endpoint] = (status, s_ver, package_schema,
                                        db_location)
        except Exception:
            LOG.error("Unable to get the status for product '%s', "
                      "considering as if the connection failed.",
                      pd.endpoint)
            prod_status[pd.endpoint] = (DBStatus.FAILED_TO_CONNECT, None,
                                        package_schema, db_location)

    return prod_status


def __db_status_check(cfg_sql_server, migration_root, environ,
                      product_name=None) -> int:
    """
    Check and print database statuses for the given product.
    """
    if not product_name:
        return 0

    LOG.debug("Checking database status for %s product.", product_name)

    prod_statuses = check_product_db_status(cfg_sql_server, migration_root,
                                            environ)

    if product_name != "all":
        avail = prod_statuses.get(product_name)
        if not avail:
            LOG.error("No product was found with this endpoint: %s",
                      str(product_name))
            return 1

        prod_statuses = {k: v for k, v in prod_statuses.items()
                         if k == product_name}

    print_prod_status(prod_statuses)
    return 0


class NonExistentProductError(Exception):
    def __init__(self, product_name):
        super().__init__(f"Non-existent product '{product_name}'")
        self.product_name = product_name


def __db_migration(migration_root,
                   environ,
                   endpoint: str,
                   connection_string: str,
                   init_instead_of_upgrade: bool) -> DBStatus:
    try:
        db = database.SQLServer.from_connection_string(connection_string,
                                                       endpoint,
                                                       RUN_META,
                                                       migration_root,
                                                       interactive=False,
                                                       env=environ)
        if init_instead_of_upgrade:
            LOG.info("[%s] Initialising...", endpoint)
            status = db.connect(init=True)
        else:
            LOG.info("[%s] Upgrading...", endpoint)
            db.connect(init=False)
            status = db.upgrade()

        status_str = database_status.db_status_msg.get(
            status, "Unknown database status")
        LOG.info("[%s] Done %s. %s", endpoint,
                 "initialising" if init_instead_of_upgrade else "upgrading",
                 status_str)
        return status
    except (CommandError, SQLAlchemyError):
        LOG.error("A database error occurred during the init/migration of "
                  "'%s'", endpoint)
        import traceback
        traceback.print_exc()
        return DBStatus.SCHEMA_INIT_ERROR if init_instead_of_upgrade \
            else DBStatus.SCHEMA_UPGRADE_FAILED
    except Exception as e:
        LOG.error("A generic error '%s' occurred during the init/migration "
                  "of '%s'", str(type(e)), endpoint)
        import traceback
        traceback.print_exc()
        return DBStatus.SCHEMA_INIT_ERROR if init_instead_of_upgrade \
            else DBStatus.SCHEMA_UPGRADE_FAILED


def __db_migration_multiple(
    cfg_sql_server, migration_root, environ,
    products_requested_for_upgrade: Optional[List[str]] = None,
    force_upgrade: bool = False
) -> int:
    """
    Migrates the schema for the product database
    ``products_requested_for_upgrade`` if specified, or all configured
    products (default).
    """
    LOG.info("Preparing schema upgrade for '%s'",
             "', '".join(products_requested_for_upgrade)
             if products_requested_for_upgrade else "<all products>")

    prod_statuses = check_product_db_status(cfg_sql_server,
                                            migration_root,
                                            environ)
    products_to_upgrade: List[str] = list()
    for endpoint in (products_requested_for_upgrade or []):
        avail = prod_statuses.get(endpoint)
        if not avail:
            LOG.error("No product was found with endpoint '%s'", endpoint)
            return 1
        products_to_upgrade.append(endpoint)
    else:
        products_to_upgrade = list(prod_statuses.keys())
    products_to_upgrade.sort()

    def _get_migration_decisions() -> List[Tuple[str, str, bool]]:
        # The lifetime of the CONFIG database connection is scoped to this
        # helper function, as keeping it alive throughout PRODUCT migrations
        # could cause timeouts.
        cfg_engine = cfg_sql_server.create_engine()
        cfg_session_factory = sessionmaker(bind=cfg_engine)
        cfg_session = cfg_session_factory()

        scheduled_upgrades_or_inits: List[Tuple[str, str, bool]] = list()
        for endpoint in products_to_upgrade:
            LOG.info("Checking: %s", endpoint)
            connection_str: Optional[str] = None

            try:
                product: Optional[ORMProduct] = cfg_session \
                    .query(ORMProduct.connection) \
                    .filter(ORMProduct.endpoint == endpoint) \
                    .one_or_none()
                if product is None:
                    raise NonExistentProductError(endpoint)

                connection_str = product.connection
            except NonExistentProductError as nepe:
                LOG.error("Attempted to upgrade product '%s', but it was not "
                          "found in the server's configuration database.",
                          nepe.product_name)
                continue
            except Exception:
                LOG.error("Failed to get the configuration for product '%s'",
                          endpoint)
                import traceback
                traceback.print_exc()
                continue

            try:
                db = database.SQLServer.from_connection_string(
                    cast(str, connection_str),
                    endpoint,
                    RUN_META,
                    migration_root,
                    interactive=False,
                    env=environ)
                db_status = db.connect()

                status_str = database_status.db_status_msg.get(
                    db_status, "Unknown database status")
                LOG.info(status_str)

                if db_status == DBStatus.SCHEMA_MISSING:
                    question = "Do you want to initialize a new schema for " \
                               f"'{endpoint}'" \
                               "? Y(es)/n(o) "
                    if force_upgrade or env.get_user_input(question):
                        LOG.info("[%s] Schema will be initialised...",
                                 endpoint)
                        scheduled_upgrades_or_inits.append(
                            (endpoint, cast(str, connection_str), True))
                    else:
                        LOG.info("[%s] No schema initialization will be done.",
                                 endpoint)
                elif db_status == DBStatus.SCHEMA_MISMATCH_OK:
                    question = f"Do you want to upgrade '{endpoint}' to new " \
                               "schema? Y(es)/n(o) "
                    if force_upgrade or env.get_user_input(question):
                        LOG.info("[%s] Schema will be upgraded...", endpoint)
                        scheduled_upgrades_or_inits.append(
                            (endpoint, cast(str, connection_str), False))
                    else:
                        LOG.info("[%s] No schema migration will be done.",
                                 endpoint)
            except (CommandError, SQLAlchemyError):
                LOG.error("A database error occurred during the preparation "
                          "for the init/migration of '%s'", endpoint)
                import traceback
                traceback.print_exc()
            except Exception as e:
                LOG.error("A generic error '%s' occurred during the "
                          "preparation for the init/migration of '%s'",
                          str(type(e)), endpoint)
                import traceback
                traceback.print_exc()

        cfg_session.close()
        cfg_engine.dispose()
        return scheduled_upgrades_or_inits

    LOG.warning("Please note after migration only newer CodeChecker versions "
                "can be used to start the server!")
    LOG.warning("It is advised to make a full backup of your run databases.")
    LOG.info("========================")
    scheduled_upgrades_or_inits = _get_migration_decisions()
    LOG.info("========================")

    if scheduled_upgrades_or_inits:
        failed_products: List[Tuple[str, DBStatus]] = list()
        thr_count = util.clamp(1, len(scheduled_upgrades_or_inits),
                               cpu_count())
        with Pool(max_workers=thr_count) as executor:
            LOG.info("Initialising/upgrading products using %d concurrent "
                     "jobs...", thr_count)
            for product_cfg, return_status in \
                    zip(scheduled_upgrades_or_inits, executor.map(
                        # Bind the first 2 non-changing arguments of
                        # __db_migration, this is fixed for the execution.
                        partial(__db_migration, migration_root, environ),
                        # Transform List[Tuple[str, str, bool]] into an
                        # Iterable[Tuple[str], Tuple[str], Tuple[bool]],
                        # and immediately unpack it, thus providing the other
                        # 3 arguments of __db_migration as a parameter pack.
                        *zip(*scheduled_upgrades_or_inits))):
                if return_status != DBStatus.OK:
                    failed_products.append((product_cfg[0], return_status))

        if failed_products:
            LOG.error("The following products failed to upgrade: %s",
                      ", ".join(list(map(lambda p: "'%s' (%s)" %
                                         (p[0],
                                          database_status.db_status_msg.get(
                                              p[1], "Unknown database status")
                                          ),
                                         failed_products))))
        else:
            LOG.info("Schema initialisation(s)/upgrade(s) executed "
                     "successfully.")
    LOG.info("========================")

    # This function always returns 0 if the upgrades were attempted, because
    # the server can start with some products that have failed to init/migrate.
    # It will just simply disallow the connection to those products.
    return 0


def kill_process_tree(parent_pid, recursive=False):
    """Stop the process tree try it gracefully first.

    Try to stop the parent and child processes gracefuly
    first if they do not stop in time send a kill signal
    to every member of the process tree.

    There is a similar function in the analyzer part please
    consider to update that in case of changing this.
    """
    proc = psutil.Process(parent_pid)
    children = proc.children(recursive)

    # Send a SIGTERM (Ctrl-C) to the main process
    proc.terminate()

    # If children processes don't stop gracefully in time,
    # slaughter them by force.
    _, still_alive = psutil.wait_procs(children, timeout=5)
    for p in still_alive:
        p.kill()

    # Wait until this process is running.
    n = 0
    timeout = 10
    while proc.is_running():
        if n > timeout:
            LOG.warning("Waiting for process %s to stop has been timed out"
                        "(timeout = %s)! Process is still running!",
                        parent_pid, timeout)
            break

        time.sleep(1)
        n += 1


def __instance_management(args):
    """Handles the instance-manager commands --list/--stop/--stop-all."""

    # TODO: The server stopping and listing must be revised on its invocation
    # once "workspace", as a concept, is removed.
    # QUESTION: What is the bestest way here to identify a server for the user?
    if 'list' in args:
        instances = instance_manager.get_instances()

        instances_on_multiple_hosts = any(True for inst in instances
                                          if inst['hostname'] !=
                                          socket.gethostname())
        if not instances_on_multiple_hosts:
            head = ['Workspace', 'View port']
        else:
            head = ['Workspace', 'Computer host', 'View port']

        rows = []
        for instance in instances:
            if not instances_on_multiple_hosts:
                rows.append((instance['workspace'], str(instance['port'])))
            else:
                rows.append((instance['workspace'],
                             instance['hostname']
                             if instance['hostname'] != socket.gethostname()
                             else '',
                             str(instance['port'])))

        print("Your running CodeChecker servers:")
        print(twodim.to_str('table', head, rows))
    elif 'stop' in args or 'stop_all' in args:
        for i in instance_manager.get_instances():
            if i['hostname'] != socket.gethostname():
                continue

            # A STOP only stops the server associated with the given workspace
            # and view-port.
            if 'stop' in args and \
                not (i['port'] == args.view_port and
                     os.path.abspath(i['workspace']) ==
                     os.path.abspath(args.config_directory)):
                continue

            try:
                kill_process_tree(i['pid'])
                LOG.info("Stopped CodeChecker server running on port %s "
                         "in workspace %s (PID: %s)",
                         i['port'], i['workspace'], i['pid'])
            except Exception:
                # Let the exception come out if the commands fail
                LOG.error("Couldn't stop process PID #%s", str(i['pid']))
                raise


def __reload_config(args):
    """
    Sends the CodeChecker server process a SIGHUP signal, causing it to
    reread it's configuration files.
    """
    for i in instance_manager.get_instances():
        if i['hostname'] != socket.gethostname():
            continue

        # A RELOAD only reloads the server associated with the given workspace
        # and view-port.
        if 'reload' in args and \
                not (i['port'] == args.view_port and
                     os.path.abspath(i['workspace']) ==
                     os.path.abspath(args.config_directory)):
            continue

        try:
            if sys.platform != "win32":
                os.kill(i['pid'], signal.SIGHUP)
        except Exception:
            LOG.error("Couldn't reload configuration file for process PID #%s",
                      str(i['pid']))
            raise


def is_localhost(address):
    """
    Check if address is one of the valid values and try to get the
    IP-addresses from the system.
    """

    valid_values = ['localhost', '0.0.0.0', '*', '::1']

    try:
        valid_values.append(socket.gethostbyname('localhost'))
    except socket.herror:
        LOG.debug("Failed to get IP address for localhost.")

    try:
        valid_values.append(socket.gethostbyname(socket.gethostname()))
    except (socket.herror, socket.gaierror):
        LOG.debug("Failed to get IP address for hostname '%s'",
                  socket.gethostname())

    return address in valid_values


def server_init_start(args):
    """
    Start or manage a CodeChecker report server.
    """

    if 'list' in args or 'stop' in args or 'stop_all' in args:
        # Set the instance manager flag to True to be able watch for it during
        # logger setup.
        setattr(args, "instance_manager", True)
        __instance_management(args)
        sys.exit(0)

    if 'reload' in args:
        __reload_config(args)
        sys.exit(0)

    # Actual server starting from this point.
    if not host_check.check_zlib():
        raise Exception("zlib is not available on the system!")

    # WARNING
    # In case of SQLite args.dbaddress default value is used
    # for which the is_localhost should return true.
    if is_localhost(args.dbaddress) and \
            not os.path.exists(args.config_directory):
        os.makedirs(args.config_directory)

    # Make sure the SQLite file can be created if it not exists.
    if 'sqlite' in args and \
            not os.path.isdir(os.path.dirname(args.sqlite)):
        os.makedirs(os.path.dirname(args.sqlite))

    if 'reset_root' in args:
        try:
            os.remove(os.path.join(args.config_directory, 'root.user'))
            LOG.info("Master superuser (root) credentials invalidated and "
                     "deleted. New ones will be generated...")
        except OSError:
            # File doesn't exist.
            pass

    if 'force_auth' in args:
        LOG.info("'--force-authentication' was passed as a command-line "
                 "option. The server will ask for users to authenticate!")

    context = webserver_context.get_context()
    context.codechecker_workspace = args.config_directory
    context.db_username = args.dbusername

    environ = env.extend(context.path_env_extra,
                         context.ld_lib_path_extra)

    cfg_sql_server = database.SQLServer.from_cmdline_args(
        vars(args), "config", CONFIG_META, context.config_migration_root,
        interactive=True, env=environ)

    LOG.info("Checking configuration database ...")
    db_status = cfg_sql_server.connect()
    db_status_msg = database_status.db_status_msg.get(db_status)
    LOG.info(db_status_msg)

    if db_status == DBStatus.SCHEMA_MISSING:
        LOG.debug("Config database schema is missing, initializing new.")
        db_status = cfg_sql_server.connect(init=True)
        if db_status != DBStatus.OK:
            LOG.error("Config database initialization failed!")
            LOG.error("Please check debug logs.")
            sys.exit(1)

    if db_status == DBStatus.SCHEMA_MISMATCH_NO:
        LOG.debug("Configuration database schema mismatch.")
        LOG.debug("No schema upgrade is possible.")
        sys.exit(1)

    force_upgrade = True if 'force_upgrade' in args else False

    if db_status == DBStatus.SCHEMA_MISMATCH_OK:
        LOG.debug("Configuration database schema mismatch!")
        LOG.debug("Schema upgrade is possible.")
        LOG.warning("Please note after migration only newer CodeChecker "
                    "versions can be used to start the server!")
        LOG.warning("It is advised to make a full backup of your "
                    "configuration database!")
        LOG.warning(cfg_sql_server.get_db_location())

        question = "Do you want to upgrade to the new schema?" \
                   " Y(es)/n(o) "
        if force_upgrade or env.get_user_input(question):
            LOG.info("Upgrading schema ...")
            new_status = cfg_sql_server.upgrade()
            status_str = database_status.db_status_msg.get(
                new_status, "Unknown database status")
            LOG.info(status_str)
            if new_status != DBStatus.OK:
                LOG.error("Schema migration failed")
                sys.exit(new_status)
        else:
            LOG.info("No schema migration was done.")
            sys.exit(0)

    if db_status == DBStatus.MISSING:
        LOG.error("Missing configuration database.")
        LOG.error("Server can not be started.")
        sys.exit(1)

    # Configuration database setup and check is needed before database
    # statuses can be checked.
    try:
        if args.status:
            ret = __db_status_check(cfg_sql_server,
                                    context.migration_root,
                                    environ,
                                    args.status)
            sys.exit(ret)
    except AttributeError:
        LOG.debug('Status was not in the arguments.')

    try:
        if args.product_to_upgrade:
            ret = __db_migration_multiple(
                cfg_sql_server,
                context.migration_root,
                environ,
                [args.product_to_upgrade]
                if args.product_to_upgrade != "all" else None,
                force_upgrade)
            sys.exit(ret)
    except AttributeError:
        LOG.debug('Product upgrade was not in the arguments.')

    # Create the main database link from the arguments passed over the
    # command line.
    cfg_dir = os.path.abspath(args.config_directory)
    default_product_path = os.path.join(cfg_dir, 'Default.sqlite')
    create_default_product = 'sqlite' in args and \
                             not os.path.exists(default_product_path)

    if create_default_product:
        # Create a default product and add it to the configuration database.
        LOG.debug("Create default product...")
        LOG.debug("Configuring schema and migration...")

        prod_server = database.SQLiteDatabase(
            "Default", default_product_path, RUN_META,
            context.run_migration_root, environ)

        LOG.debug("Checking 'Default' product database.")
        db_status = prod_server.connect()
        if db_status != DBStatus.MISSING:
            db_status = prod_server.connect(init=True)
            LOG.debug(database_status.db_status_msg.get(db_status))
            if db_status != DBStatus.OK:
                LOG.error("Failed to configure default product")
                sys.exit(1)

        product_conn_string = prod_server.get_connection_string()

        server.add_initial_run_database(
            cfg_sql_server, product_conn_string)

        LOG.info("Product 'Default' at '%s' created and set up.",
                 default_product_path)

    prod_statuses = check_product_db_status(cfg_sql_server,
                                            context.run_migration_root,
                                            environ)

    upgrade_available = {}
    for k, v in prod_statuses.items():
        db_status, _, _, _ = v
        if db_status == DBStatus.SCHEMA_MISMATCH_OK or \
                db_status == DBStatus.SCHEMA_MISSING:
            upgrade_available[k] = v

    if upgrade_available:
        print_prod_status(prod_statuses)
        LOG.warning("Multiple products can be upgraded, make a backup!")
        __db_migration_multiple(cfg_sql_server,
                                context.run_migration_root,
                                environ,
                                None,
                                force_upgrade)

    prod_statuses = check_product_db_status(cfg_sql_server,
                                            context.run_migration_root,
                                            environ)
    print_prod_status(prod_statuses)

    non_ok_db = False
    for k, v in prod_statuses.items():
        db_status, _, _, _ = v
        if db_status != DBStatus.OK:
            non_ok_db = True
        break

    if non_ok_db:
        LOG.error("There are some database issues.")
        if not force_upgrade:
            status_str = "Do you want to start the server? Y(es)/n(o) "
            if not env.get_user_input(status_str):
                sys.exit(1)

    # Start database viewer.
    package_data = {'www_root': context.www_root,
                    'doc_root': context.doc_root,
                    'version': context.package_git_tag}

    try:
        server.start_server(args.config_directory,
                            package_data,
                            args.view_port,
                            cfg_sql_server,
                            args.listen_address,
                            'force_auth' in args,
                            args.skip_db_cleanup,
                            context,
                            environ)
    except socket.error as err:
        if err.errno == errno.EADDRINUSE:
            LOG.error("Server can't be started, maybe port number (%s) is "
                      "already used. Check the connection parameters. Use "
                      "the option '-p 0' to find a free port automatically.",
                      args.view_port)
            sys.exit(1)
        else:
            raise


def main(args):
    """
    Setup a logger server based on the configuration and
    manage the CodeChecker server.
    """
    workspace = (
        args.config_directory
        if "config_directory" in args and not hasattr(args, "instance_manager")
        else None
    )

    # Create workspace directory before logging is initialized.
    if workspace and not os.path.exists(args.config_directory):
        LOG.info("Creating non existing config directory: %s",
                 args.config_directory)
        os.makedirs(args.config_directory)

    with logger.LOG_CFG_SERVER(
        args.verbose if "verbose" in args else None, workspace=workspace
    ):
        try:
            cmd_config.check_config_file(args)
        except FileNotFoundError as fnerr:
            LOG.error(fnerr)
            sys.exit(1)
        server_init_start(args)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Instance manager handles the state keeping of running CodeChecker instances
for a particular user on the local machine.
"""


import getpass
import json
import os
import psutil
import socket
import stat

import portalocker

from codechecker_common.logger import get_logger
from codechecker_common.util import load_json

LOG = get_logger('system')


def __get_instance_descriptor_path(folder=None):
    if not folder:
        folder = os.path.expanduser("~")

    return os.path.join(folder, ".codechecker.instances.json")


def __make_instance_descriptor_file(folder=None):
    descriptor = __get_instance_descriptor_path(folder)
    if not os.path.exists(descriptor):
        with open(descriptor, 'w',
                  encoding="utf-8", errors="ignore") as f:
            json.dump([], f)
        os.chmod(descriptor, stat.S_IRUSR | stat.S_IWUSR)


def __check_instance(hostname, pid):
    """Check if the given process on the system is a valid, running CodeChecker
    for the current user."""

    # Instances running on a remote host with a filesystem shared with us can
    # not usually be checked (/proc is rarely shared across computers...),
    # so we consider them "alive" servers.
    if hostname != socket.gethostname():
        return True

    try:
        proc = psutil.Process(pid)

        cli = os.path.join("codechecker_common", "cli.py")
        return cli in proc.cmdline()[1] and \
            proc.username() == getpass.getuser()
    except psutil.NoSuchProcess:
        # If the process does not exist, it cannot be valid.
        return False


def __rewrite_instance_file(append, remove, folder=None):
    """
    This helper method reads the user's instance descriptor and manages it
    eliminating dead records, appending new ones and re-serialising the file.
    """
    __make_instance_descriptor_file(folder)

    append_pids = [i['pid'] for i in append]

    instance_descriptor_file = __get_instance_descriptor_path(folder)
    with open(instance_descriptor_file, 'r+',
              encoding="utf-8", errors="ignore") as instance_file:
        portalocker.lock(instance_file, portalocker.LOCK_EX)

        instances = []
        try:
            instances = json.loads(instance_file.read())
        except (ValueError, TypeError) as ex:
            LOG.warning('Failed to process json file: %s',
                        instance_descriptor_file)
            LOG.warning(ex)

        # After reading, check every instance if they are still valid and
        # make sure PID does not collide accidentally with the
        # to-be-registered instances, if any exists in the append list as it
        # would cause duplication.
        #
        # Also, we remove the records to the given PIDs, if any exists.
        instances = [i for i in instances
                     if i['pid'] not in append_pids and
                     (i['hostname'] + ":" + str(i['pid'])) not in remove]

        instances = instances + append

        instance_file.seek(0)
        instance_file.truncate()
        json.dump(instances, instance_file, indent=2)
        portalocker.unlock(instance_file)


def register(pid, workspace, port, folder=None):
    """
    Adds the specified CodeChecker server instance to the user's instance
    descriptor.
    """

    __rewrite_instance_file([{"pid": pid,
                              "hostname": socket.gethostname(),
                              "workspace": workspace,
                              "port": port}],
                            [],
                            folder)


def unregister(pid, folder=None):
    """
    Removes the specified CodeChecker server instance from the user's instance
    descriptor.
    """

    __rewrite_instance_file([],
                            [socket.gethostname() + ":" + str(pid)],
                            folder)


def get_instances(folder=None):
    """Returns the list of running servers for the current user."""

    # This method does NOT write the descriptor file.

    descriptor = __get_instance_descriptor_path(folder)
    instances = load_json(descriptor, {}, lock=True)

    return [i for i in instances if __check_instance(i['hostname'], i['pid'])]



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import logging
from .ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
all_structs = []


class Iface(object):
    def getPackageVersion(self):
        pass

    def isAdministratorOfAnyProduct(self):
        pass

    def getProducts(self, productEndpointFilter, productNameFilter):
        """
        Parameters:
         - productEndpointFilter
         - productNameFilter
        """
        pass

    def getCurrentProduct(self):
        pass

    def getProductConfiguration(self, productId):
        """
        Parameters:
         - productId
        """
        pass

    def addProduct(self, product):
        """
        Parameters:
         - product
        """
        pass

    def editProduct(self, productId, newConfiguration):
        """
        Parameters:
         - productId
         - newConfiguration
        """
        pass

    def removeProduct(self, productId):
        """
        Parameters:
         - productId
        """
        pass


class Client(Iface):
    def __init__(self, iprot, oprot=None):
        self._iprot = self._oprot = iprot
        if oprot is not None:
            self._oprot = oprot
        self._seqid = 0

    def getPackageVersion(self):
        self.send_getPackageVersion()
        return self.recv_getPackageVersion()

    def send_getPackageVersion(self):
        self._oprot.writeMessageBegin('getPackageVersion', TMessageType.CALL, self._seqid)
        args = getPackageVersion_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getPackageVersion(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getPackageVersion_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getPackageVersion failed: unknown result")

    def isAdministratorOfAnyProduct(self):
        self.send_isAdministratorOfAnyProduct()
        return self.recv_isAdministratorOfAnyProduct()

    def send_isAdministratorOfAnyProduct(self):
        self._oprot.writeMessageBegin('isAdministratorOfAnyProduct', TMessageType.CALL, self._seqid)
        args = isAdministratorOfAnyProduct_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_isAdministratorOfAnyProduct(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = isAdministratorOfAnyProduct_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.requestError is not None:
            raise result.requestError
        raise TApplicationException(TApplicationException.MISSING_RESULT, "isAdministratorOfAnyProduct failed: unknown result")

    def getProducts(self, productEndpointFilter, productNameFilter):
        """
        Parameters:
         - productEndpointFilter
         - productNameFilter
        """
        self.send_getProducts(productEndpointFilter, productNameFilter)
        return self.recv_getProducts()

    def send_getProducts(self, productEndpointFilter, productNameFilter):
        self._oprot.writeMessageBegin('getProducts', TMessageType.CALL, self._seqid)
        args = getProducts_args()
        args.productEndpointFilter = productEndpointFilter
        args.productNameFilter = productNameFilter
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getProducts(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getProducts_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.requestError is not None:
            raise result.requestError
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getProducts failed: unknown result")

    def getCurrentProduct(self):
        self.send_getCurrentProduct()
        return self.recv_getCurrentProduct()

    def send_getCurrentProduct(self):
        self._oprot.writeMessageBegin('getCurrentProduct', TMessageType.CALL, self._seqid)
        args = getCurrentProduct_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getCurrentProduct(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getCurrentProduct_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.requestError is not None:
            raise result.requestError
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getCurrentProduct failed: unknown result")

    def getProductConfiguration(self, productId):
        """
        Parameters:
         - productId
        """
        self.send_getProductConfiguration(productId)
        return self.recv_getProductConfiguration()

    def send_getProductConfiguration(self, productId):
        self._oprot.writeMessageBegin('getProductConfiguration', TMessageType.CALL, self._seqid)
        args = getProductConfiguration_args()
        args.productId = productId
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getProductConfiguration(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getProductConfiguration_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.requestError is not None:
            raise result.requestError
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getProductConfiguration failed: unknown result")

    def addProduct(self, product):
        """
        Parameters:
         - product
        """
        self.send_addProduct(product)
        return self.recv_addProduct()

    def send_addProduct(self, product):
        self._oprot.writeMessageBegin('addProduct', TMessageType.CALL, self._seqid)
        args = addProduct_args()
        args.product = product
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_addProduct(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = addProduct_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.requestError is not None:
            raise result.requestError
        raise TApplicationException(TApplicationException.MISSING_RESULT, "addProduct failed: unknown result")

    def editProduct(self, productId, newConfiguration):
        """
        Parameters:
         - productId
         - newConfiguration
        """
        self.send_editProduct(productId, newConfiguration)
        return self.recv_editProduct()

    def send_editProduct(self, productId, newConfiguration):
        self._oprot.writeMessageBegin('editProduct', TMessageType.CALL, self._seqid)
        args = editProduct_args()
        args.productId = productId
        args.newConfiguration = newConfiguration
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_editProduct(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = editProduct_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.requestError is not None:
            raise result.requestError
        raise TApplicationException(TApplicationException.MISSING_RESULT, "editProduct failed: unknown result")

    def removeProduct(self, productId):
        """
        Parameters:
         - productId
        """
        self.send_removeProduct(productId)
        return self.recv_removeProduct()

    def send_removeProduct(self, productId):
        self._oprot.writeMessageBegin('removeProduct', TMessageType.CALL, self._seqid)
        args = removeProduct_args()
        args.productId = productId
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_removeProduct(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = removeProduct_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        if result.requestError is not None:
            raise result.requestError
        raise TApplicationException(TApplicationException.MISSING_RESULT, "removeProduct failed: unknown result")


class Processor(Iface, TProcessor):
    def __init__(self, handler):
        self._handler = handler
        self._processMap = {}
        self._processMap["getPackageVersion"] = Processor.process_getPackageVersion
        self._processMap["isAdministratorOfAnyProduct"] = Processor.process_isAdministratorOfAnyProduct
        self._processMap["getProducts"] = Processor.process_getProducts
        self._processMap["getCurrentProduct"] = Processor.process_getCurrentProduct
        self._processMap["getProductConfiguration"] = Processor.process_getProductConfiguration
        self._processMap["addProduct"] = Processor.process_addProduct
        self._processMap["editProduct"] = Processor.process_editProduct
        self._processMap["removeProduct"] = Processor.process_removeProduct

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_getPackageVersion(self, seqid, iprot, oprot):
        args = getPackageVersion_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getPackageVersion_result()
        try:
            result.success = self._handler.getPackageVersion()
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getPackageVersion", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_isAdministratorOfAnyProduct(self, seqid, iprot, oprot):
        args = isAdministratorOfAnyProduct_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = isAdministratorOfAnyProduct_result()
        try:
            result.success = self._handler.isAdministratorOfAnyProduct()
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except codechecker_api_shared.ttypes.RequestFailed as requestError:
            msg_type = TMessageType.REPLY
            result.requestError = requestError
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("isAdministratorOfAnyProduct", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getProducts(self, seqid, iprot, oprot):
        args = getProducts_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getProducts_result()
        try:
            result.success = self._handler.getProducts(args.productEndpointFilter, args.productNameFilter)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except codechecker_api_shared.ttypes.RequestFailed as requestError:
            msg_type = TMessageType.REPLY
            result.requestError = requestError
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getProducts", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getCurrentProduct(self, seqid, iprot, oprot):
        args = getCurrentProduct_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getCurrentProduct_result()
        try:
            result.success = self._handler.getCurrentProduct()
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except codechecker_api_shared.ttypes.RequestFailed as requestError:
            msg_type = TMessageType.REPLY
            result.requestError = requestError
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getCurrentProduct", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getProductConfiguration(self, seqid, iprot, oprot):
        args = getProductConfiguration_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getProductConfiguration_result()
        try:
            result.success = self._handler.getProductConfiguration(args.productId)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except codechecker_api_shared.ttypes.RequestFailed as requestError:
            msg_type = TMessageType.REPLY
            result.requestError = requestError
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getProductConfiguration", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_addProduct(self, seqid, iprot, oprot):
        args = addProduct_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = addProduct_result()
        try:
            result.success = self._handler.addProduct(args.product)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except codechecker_api_shared.ttypes.RequestFailed as requestError:
            msg_type = TMessageType.REPLY
            result.requestError = requestError
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("addProduct", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_editProduct(self, seqid, iprot, oprot):
        args = editProduct_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = editProduct_result()
        try:
            result.success = self._handler.editProduct(args.productId, args.newConfiguration)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except codechecker_api_shared.ttypes.RequestFailed as requestError:
            msg_type = TMessageType.REPLY
            result.requestError = requestError
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("editProduct", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_removeProduct(self, seqid, iprot, oprot):
        args = removeProduct_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = removeProduct_result()
        try:
            result.success = self._handler.removeProduct(args.productId)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except codechecker_api_shared.ttypes.RequestFailed as requestError:
            msg_type = TMessageType.REPLY
            result.requestError = requestError
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("removeProduct", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

# HELPER FUNCTIONS AND STRUCTURES


class getPackageVersion_args(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getPackageVersion_args')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getPackageVersion_args)
getPackageVersion_args.thrift_spec = (
)


class getPackageVersion_result(object):
    """
    Attributes:
     - success
    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRING:
                    self.success = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getPackageVersion_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRING, 0)
            oprot.writeString(self.success.encode('utf-8') if sys.version_info[0] == 2 else self.success)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getPackageVersion_result)
getPackageVersion_result.thrift_spec = (
    (0, TType.STRING, 'success', 'UTF8', None, ),  # 0
)


class isAdministratorOfAnyProduct_args(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('isAdministratorOfAnyProduct_args')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(isAdministratorOfAnyProduct_args)
isAdministratorOfAnyProduct_args.thrift_spec = (
)


class isAdministratorOfAnyProduct_result(object):
    """
    Attributes:
     - success
     - requestError
    """


    def __init__(self, success=None, requestError=None,):
        self.success = success
        self.requestError = requestError

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.BOOL:
                    self.success = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.requestError = codechecker_api_shared.ttypes.RequestFailed()
                    self.requestError.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('isAdministratorOfAnyProduct_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.BOOL, 0)
            oprot.writeBool(self.success)
            oprot.writeFieldEnd()
        if self.requestError is not None:
            oprot.writeFieldBegin('requestError', TType.STRUCT, 1)
            self.requestError.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(isAdministratorOfAnyProduct_result)
isAdministratorOfAnyProduct_result.thrift_spec = (
    (0, TType.BOOL, 'success', None, None, ),  # 0
    (1, TType.STRUCT, 'requestError', [codechecker_api_shared.ttypes.RequestFailed, None], None, ),  # 1
)


class getProducts_args(object):
    """
    Attributes:
     - productEndpointFilter
     - productNameFilter
    """


    def __init__(self, productEndpointFilter=None, productNameFilter=None,):
        self.productEndpointFilter = productEndpointFilter
        self.productNameFilter = productNameFilter

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.productEndpointFilter = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.productNameFilter = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProducts_args')
        if self.productEndpointFilter is not None:
            oprot.writeFieldBegin('productEndpointFilter', TType.STRING, 1)
            oprot.writeString(self.productEndpointFilter.encode('utf-8') if sys.version_info[0] == 2 else self.productEndpointFilter)
            oprot.writeFieldEnd()
        if self.productNameFilter is not None:
            oprot.writeFieldBegin('productNameFilter', TType.STRING, 2)
            oprot.writeString(self.productNameFilter.encode('utf-8') if sys.version_info[0] == 2 else self.productNameFilter)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProducts_args)
getProducts_args.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'productEndpointFilter', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'productNameFilter', 'UTF8', None, ),  # 2
)


class getProducts_result(object):
    """
    Attributes:
     - success
     - requestError
    """


    def __init__(self, success=None, requestError=None,):
        self.success = success
        self.requestError = requestError

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.LIST:
                    self.success = []
                    (_etype17, _size14) = iprot.readListBegin()
                    for _i18 in range(_size14):
                        _elem19 = Product()
                        _elem19.read(iprot)
                        self.success.append(_elem19)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.requestError = codechecker_api_shared.ttypes.RequestFailed()
                    self.requestError.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProducts_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.LIST, 0)
            oprot.writeListBegin(TType.STRUCT, len(self.success))
            for iter20 in self.success:
                iter20.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.requestError is not None:
            oprot.writeFieldBegin('requestError', TType.STRUCT, 1)
            self.requestError.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProducts_result)
getProducts_result.thrift_spec = (
    (0, TType.LIST, 'success', (TType.STRUCT, [Product, None], False), None, ),  # 0
    (1, TType.STRUCT, 'requestError', [codechecker_api_shared.ttypes.RequestFailed, None], None, ),  # 1
)


class getCurrentProduct_args(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getCurrentProduct_args')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getCurrentProduct_args)
getCurrentProduct_args.thrift_spec = (
)


class getCurrentProduct_result(object):
    """
    Attributes:
     - success
     - requestError
    """


    def __init__(self, success=None, requestError=None,):
        self.success = success
        self.requestError = requestError

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = Product()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.requestError = codechecker_api_shared.ttypes.RequestFailed()
                    self.requestError.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getCurrentProduct_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        if self.requestError is not None:
            oprot.writeFieldBegin('requestError', TType.STRUCT, 1)
            self.requestError.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getCurrentProduct_result)
getCurrentProduct_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [Product, None], None, ),  # 0
    (1, TType.STRUCT, 'requestError', [codechecker_api_shared.ttypes.RequestFailed, None], None, ),  # 1
)


class getProductConfiguration_args(object):
    """
    Attributes:
     - productId
    """


    def __init__(self, productId=None,):
        self.productId = productId

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I64:
                    self.productId = iprot.readI64()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProductConfiguration_args')
        if self.productId is not None:
            oprot.writeFieldBegin('productId', TType.I64, 1)
            oprot.writeI64(self.productId)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProductConfiguration_args)
getProductConfiguration_args.thrift_spec = (
    None,  # 0
    (1, TType.I64, 'productId', None, None, ),  # 1
)


class getProductConfiguration_result(object):
    """
    Attributes:
     - success
     - requestError
    """


    def __init__(self, success=None, requestError=None,):
        self.success = success
        self.requestError = requestError

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = ProductConfiguration()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.requestError = codechecker_api_shared.ttypes.RequestFailed()
                    self.requestError.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProductConfiguration_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        if self.requestError is not None:
            oprot.writeFieldBegin('requestError', TType.STRUCT, 1)
            self.requestError.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProductConfiguration_result)
getProductConfiguration_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [ProductConfiguration, None], None, ),  # 0
    (1, TType.STRUCT, 'requestError', [codechecker_api_shared.ttypes.RequestFailed, None], None, ),  # 1
)


class addProduct_args(object):
    """
    Attributes:
     - product
    """


    def __init__(self, product=None,):
        self.product = product

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.product = ProductConfiguration()
                    self.product.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('addProduct_args')
        if self.product is not None:
            oprot.writeFieldBegin('product', TType.STRUCT, 1)
            self.product.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(addProduct_args)
addProduct_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'product', [ProductConfiguration, None], None, ),  # 1
)


class addProduct_result(object):
    """
    Attributes:
     - success
     - requestError
    """


    def __init__(self, success=None, requestError=None,):
        self.success = success
        self.requestError = requestError

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.BOOL:
                    self.success = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.requestError = codechecker_api_shared.ttypes.RequestFailed()
                    self.requestError.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('addProduct_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.BOOL, 0)
            oprot.writeBool(self.success)
            oprot.writeFieldEnd()
        if self.requestError is not None:
            oprot.writeFieldBegin('requestError', TType.STRUCT, 1)
            self.requestError.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(addProduct_result)
addProduct_result.thrift_spec = (
    (0, TType.BOOL, 'success', None, None, ),  # 0
    (1, TType.STRUCT, 'requestError', [codechecker_api_shared.ttypes.RequestFailed, None], None, ),  # 1
)


class editProduct_args(object):
    """
    Attributes:
     - productId
     - newConfiguration
    """


    def __init__(self, productId=None, newConfiguration=None,):
        self.productId = productId
        self.newConfiguration = newConfiguration

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I64:
                    self.productId = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.newConfiguration = ProductConfiguration()
                    self.newConfiguration.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('editProduct_args')
        if self.productId is not None:
            oprot.writeFieldBegin('productId', TType.I64, 1)
            oprot.writeI64(self.productId)
            oprot.writeFieldEnd()
        if self.newConfiguration is not None:
            oprot.writeFieldBegin('newConfiguration', TType.STRUCT, 2)
            self.newConfiguration.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(editProduct_args)
editProduct_args.thrift_spec = (
    None,  # 0
    (1, TType.I64, 'productId', None, None, ),  # 1
    (2, TType.STRUCT, 'newConfiguration', [ProductConfiguration, None], None, ),  # 2
)


class editProduct_result(object):
    """
    Attributes:
     - success
     - requestError
    """


    def __init__(self, success=None, requestError=None,):
        self.success = success
        self.requestError = requestError

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.BOOL:
                    self.success = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.requestError = codechecker_api_shared.ttypes.RequestFailed()
                    self.requestError.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('editProduct_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.BOOL, 0)
            oprot.writeBool(self.success)
            oprot.writeFieldEnd()
        if self.requestError is not None:
            oprot.writeFieldBegin('requestError', TType.STRUCT, 1)
            self.requestError.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(editProduct_result)
editProduct_result.thrift_spec = (
    (0, TType.BOOL, 'success', None, None, ),  # 0
    (1, TType.STRUCT, 'requestError', [codechecker_api_shared.ttypes.RequestFailed, None], None, ),  # 1
)


class removeProduct_args(object):
    """
    Attributes:
     - productId
    """


    def __init__(self, productId=None,):
        self.productId = productId

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I64:
                    self.productId = iprot.readI64()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('removeProduct_args')
        if self.productId is not None:
            oprot.writeFieldBegin('productId', TType.I64, 1)
            oprot.writeI64(self.productId)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(removeProduct_args)
removeProduct_args.thrift_spec = (
    None,  # 0
    (1, TType.I64, 'productId', None, None, ),  # 1
)


class removeProduct_result(object):
    """
    Attributes:
     - success
     - requestError
    """


    def __init__(self, success=None, requestError=None,):
        self.success = success
        self.requestError = requestError

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.BOOL:
                    self.success = iprot.readBool()
                else:
                    iprot.skip(ftype)
            elif fid == 1:
                if ftype == TType.STRUCT:
                    self.requestError = codechecker_api_shared.ttypes.RequestFailed()
                    self.requestError.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('removeProduct_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.BOOL, 0)
            oprot.writeBool(self.success)
            oprot.writeFieldEnd()
        if self.requestError is not None:
            oprot.writeFieldBegin('requestError', TType.STRUCT, 1)
            self.requestError.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(removeProduct_result)
removeProduct_result.thrift_spec = (
    (0, TType.BOOL, 'success', None, None, ),  # 0
    (1, TType.STRUCT, 'requestError', [codechecker_api_shared.ttypes.RequestFailed, None], None, ),  # 1
)
fix_spec(all_structs)
del all_structs




# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import json
import logging
import os
import sys
import fnmatch
import re

from typing import Dict, List, Optional, TextIO


LOG = logging.getLogger('report-converter')


def get_last_mod_time(file_path: str) -> Optional[float]:
    """ Return the last modification time of a file. """
    try:
        return os.stat(file_path).st_mtime
    except OSError as err:
        LOG.debug("File is missing: %s", err)
        return None


def get_linef(fp: TextIO, line_no: int) -> str:
    """'fp' should be (readable) file object.
    Return the line content at line_no or an empty line
    if there is less lines than line_no.
    """
    fp.seek(0)
    for line in fp:
        line_no -= 1
        if line_no == 0:
            return line
    return ''


def get_line(file_path: str, line_no: int, errors: str = 'ignore') -> str:
    """
    Return the given line from the file. If line_no is larger than the number
    of lines in the file then empty string returns.
    If the file can't be opened for read, the function also returns empty
    string.

    Try to encode every file as utf-8 to read the line content do not depend
    on the platform settings. By default locale.getpreferredencoding() is used
    which depends on the platform.

    Changing the encoding error handling can influence the hash content!
    """
    try:
        with open(file_path, mode='r', encoding='utf-8', errors=errors) as f:
            return get_linef(f, line_no)
    except IOError:
        LOG.error("Failed to open file %s", file_path)
        return ''


def trim_path_prefixes(path: str, prefixes: Optional[List[str]]) -> str:
    """
    Removes the longest matching leading path from the file path.
    """

    # If no prefixes are specified.
    if not prefixes:
        return path

    # Find the longest matching prefix in the path.
    longest_matching_prefix = None
    for prefix in prefixes:
        if not prefix.endswith('/'):
            prefix += '/'

        regex_str = fnmatch.translate(prefix)
        assert regex_str[-2:] == '\\Z', \
               r'fnmatch.translate should leave \\Z at the end of the matcher!'

        prefix_matcher = re.compile(regex_str[:-2])

        matches = prefix_matcher.match(path)
        if matches:
            matching_prefix = matches[0]
            if not longest_matching_prefix or \
                    len(longest_matching_prefix) < len(matching_prefix):

                longest_matching_prefix = matching_prefix

    # If no prefix found or the longest prefix is the root do not trim the
    # path.
    if not longest_matching_prefix or longest_matching_prefix == '/':
        return path

    return path[len(longest_matching_prefix):]


def dump_json_output(
    data: Dict,
    output_file_path: Optional[str] = None,
    out=sys.stdout
) -> str:
    """
    Write JSON data to the given output file and returns the written output.
    """
    data_str = json.dumps(data)

    # Write output data to the file if given.
    if output_file_path:
        with open(output_file_path, mode='w',
                  encoding='utf-8', errors="ignore") as f:
            f.write(data_str)

        LOG.info('JSON report file was created: %s', output_file_path)
    elif out:
        out.write(f"{data_str}\n")

    return data_str
